# namedtuple is needed for find_mod_objs so it can have a non-local module

import sys
from collections import namedtuple

import pytest

from ..utils import find_mod_objs

PY3 = sys.version_info[0] >= 3
pytestmark = pytest.mark.skipif("PY3")


def test_find_mod_objs():
    lnms, fqns, objs = find_mod_objs('astropy_helpers')

    # this import  is after the above call intentionally to make sure
    # find_mod_objs properly imports astropy on its own
    import astropy_helpers

    # just check for astropy.test ... other things might be added, so we
    # shouldn't check that it's the only thing
    assert lnms == []

    lnms, fqns, objs = find_mod_objs(
        'astropy_helpers.sphinx.ext.tests.test_utils', onlylocals=False)

    assert namedtuple in objs

    lnms, fqns, objs = find_mod_objs(
        'astropy_helpers.sphinx.ext.tests.test_utils', onlylocals=True)
    assert 'namedtuple' not in lnms
    assert 'collections.namedtuple' not in fqns
    assert namedtuple not in objs



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
Module to remove cosmic rays from an astronomical image using the
L.A.Cosmic (PASP 113, 1420, 2001) algorithm.
"""

from __future__ import (absolute_import, division, print_function,
                        unicode_literals)
from distutils.version import LooseVersion
import numpy as np

import astropy
if LooseVersion(astropy.__version__) < LooseVersion('1.1'):
    raise ImportError('lacosmic requires astropy >= 1.1')

from astropy.nddata.utils import block_reduce, block_replicate
from astropy import log


__all__ = ['lacosmic']


def lacosmic(data, contrast, cr_threshold, neighbor_threshold,
             error=None, mask=None, background=None, effective_gain=None,
             readnoise=None, maxiter=4, border_mode='mirror'):
    """
    Remove cosmic rays from an astronomical image using the `L.A.Cosmic
    <http://www.astro.yale.edu/dokkum/lacosmic/>`_ algorithm.  The
    algorithm is based on Laplacian edge detection and is described in
    `PASP 113, 1420 (2001)`_.

    .. _PASP 113, 1420 (2001):
        http://adsabs.harvard.edu/abs/2001PASP..113.1420V

    Parameters
    ----------
    data : array_like
        The 2D array of the image.

    contrast : float
        Contrast threshold between the Laplacian image and the
        fine-structure image.  If your image is critically sampled, use
        a value around 2.  If your image is undersampled (e.g. HST
        data), a value of 4 or 5 (or more) is more appropriate.  If your
        image is oversampled, use a value between 1 and 2.  For details,
        please see `PASP 113, 1420 (2001)`_, which calls this parameter
        :math:`f_{\mbox{lim}}`.  In particular, Figure 4 shows the
        approximate relationship between the ``contrast`` parameter and
        the pixel full-width half-maximum of stars in your image.

    cr_threshold : float
        The Laplacian signal-to-noise ratio threshold for cosmic-ray
        detection.

    neighbor_threshold : float
        The Laplacian signal-to-noise ratio threshold for detection of
        cosmic rays in pixels neighboring the initially-identified
        cosmic rays.

    error : array_like, optional
        The pixel-wise Gaussian 1-sigma errors of the input ``data``.
        If ``error`` is not input, then ``effective_gain`` and
        ``readnoise`` will be used to construct an approximate model of
        the ``error``.  If ``error`` is input, it will override the
        ``effective_gain`` and ``readnoise`` parameters.  ``error`` must
        have the same shape as ``data``.

    mask : array_like (bool), optional
        A boolean mask, with the same shape as ``data``, where a `True`
        value indicates the corresponding element of ``data`` is masked.
        Masked pixels are ignored when identifying cosmic rays.  It is
        highly recommended that saturated stars be included in ``mask``.

    background : float or array_like, optional
        The background level previously subtracted from the input
        ``data``.  ``background`` may either be a scalar value or a 2D
        image with the same shape as the input ``data``.  If the input
        ``data`` has not been background-subtracted, then set
        ``background=None`` (default).

    effective_gain : float, array-like, optional
        Ratio of counts (e.g., electrons or photons) to the units of
        ``data``.  For example, if your input ``data`` are in units of
        ADU, then ``effective_gain`` should represent electrons/ADU.  If
        your input ``data`` are in units of electrons/s then
        ``effective_gain`` should be the exposure time (or an exposure
        time map).  ``effective_gain`` and ``readnoise`` must be
        specified if ``error`` is not input.

    readnoise : float, optional
        The read noise (in electrons) in the input ``data``.
        ``effective_gain`` and ``readnoise`` must be specified if
        ``error`` is not input.

    maxiter : float, optional
        The maximum number of iterations.  The default is 4.  The
        routine will automatically exit if no additional cosmic rays are
        identified.  If the routine is still identifying cosmic rays
        after four iterations, then you are likely digging into sources
        (e.g. saturated stars) and/or the noise.  In that case, try
        inputing a ``mask`` or increasing the value of ``cr_threshold``.

    border_mode : {'reflect', 'constant', 'nearest', 'mirror', 'wrap'}, optional
        The mode in which the array borders are handled during
        convolution and median filtering.  For 'constant', the value is
        0.  The default is 'mirror', which matches the original
        L.A.Cosmic algorithm.

    Returns
    -------
    cleaned_image : `~numpy.ndarray`
        The cosmic-ray cleaned image.

    crmask : `~numpy.ndarray` (bool)
        A mask image of the identified cosmic rays.  Cosmic-ray pixels
        have a value of `True`.
    """

    from scipy import ndimage
    block_size = 2.0
    kernel = np.array([[0.0, -1.0, 0.0], [-1.0, 4.0, -1.0], [0.0, -1.0, 0.0]])

    clean_data = data.copy()
    if background is not None:
        clean_data += background
    final_crmask = np.zeros(data.shape, dtype=bool)

    if error is not None:
        if data.shape != error.shape:
            raise ValueError('error and data must have the same shape')
    clean_error_image = error

    ncosmics, ncosmics_tot = 0, 0
    for iteration in range(maxiter):
        sampled_img = block_replicate(clean_data, block_size)
        convolved_img = ndimage.convolve(sampled_img, kernel,
                                         mode=border_mode).clip(min=0.0)
        laplacian_img = block_reduce(convolved_img, block_size)

        if clean_error_image is None:
            if effective_gain is None or readnoise is None:
                raise AssertionError('effective_gain and readnoise must be '
                                     'input if error is not input')
            med5_img = ndimage.median_filter(clean_data, size=5,
                                             mode=border_mode).clip(min=1.e-5)
            error_image = (np.sqrt(effective_gain*med5_img + readnoise**2) /
                           effective_gain)
        else:
            error_image = clean_error_image

        snr_img = laplacian_img / (block_size * error_image)
        # this is used to remove extended structures (larger than ~5x5)
        snr_img -= ndimage.median_filter(snr_img, size=5, mode=border_mode)

        # used to remove compact bright objects
        med3_img = ndimage.median_filter(clean_data, size=3, mode=border_mode)
        med7_img = ndimage.median_filter(med3_img, size=7, mode=border_mode)
        finestruct_img = ((med3_img - med7_img) / error_image).clip(min=0.01)

        cr_mask1 = snr_img > cr_threshold
        # NOTE: to follow the paper exactly, this condition should be
        # "> contrast * block_size".  "lacos_im.cl" uses simply "> contrast"
        cr_mask2 = (snr_img / finestruct_img) > contrast
        cr_mask = cr_mask1 * cr_mask2
        if mask is not None:
            cr_mask = np.logical_and(cr_mask, ~mask)

        # grow cosmic rays by one pixel and check in snr_img
        selem = np.ones((3, 3))
        neigh_mask = ndimage.binary_dilation(cr_mask, selem)
        cr_mask = cr_mask1 * neigh_mask
        # now grow one more pixel and lower the detection threshold
        neigh_mask = ndimage.binary_dilation(cr_mask, selem)
        cr_mask = (snr_img > neighbor_threshold) * neigh_mask

        # previously unknown cosmic rays found in this iteration
        crmask_new = np.logical_and(~final_crmask, cr_mask)
        ncosmics = np.count_nonzero(crmask_new)

        final_crmask = np.logical_or(final_crmask, cr_mask)
        ncosmics_tot += ncosmics
        log.info('Iteration {0}: Found {1} cosmic-ray pixels, '
                 'Total: {2}'.format(iteration + 1, ncosmics, ncosmics_tot))
        if ncosmics == 0:
            if background is not None:
                clean_data -= background
            return clean_data, final_crmask
        clean_data = _clean_masked_pixels(clean_data, final_crmask, size=5,
                                          exclude_mask=mask)

    if background is not None:
        clean_data -= background
    return clean_data, final_crmask


def _clean_masked_pixels(data, mask, size=5, exclude_mask=None):
    """
    Clean masked pixels in an image.  Each masked pixel is replaced by
    the median of unmasked pixels in a 2D window of ``size`` centered on
    it.  If all pixels in the window are masked, then the window is
    increased in size until unmasked pixels are found.

    Pixels in ``exclude_mask`` are not cleaned, but they are excluded
    when calculating the local median.
    """

    assert size % 2 == 1, 'size must be an odd integer'
    assert data.shape == mask.shape, \
        'mask must have the same shape as image'
    ny, nx = data.shape
    mask_coords = np.argwhere(mask)
    if exclude_mask is not None:
        assert data.shape == exclude_mask.shape, \
            'exclude_mask must have the same shape as data'
        maskall = np.logical_or(mask, exclude_mask)
    else:
        maskall = mask
    mask_idx = maskall.nonzero()
    data_nanmask = data.copy()
    data_nanmask[mask_idx] = np.nan

    nexpanded = 0
    for coord in mask_coords:
        y, x = coord
        median_val, expanded = _local_median(data_nanmask, x, y, nx, ny,
                                             size=size)
        data[y, x] = median_val
        if expanded:
            nexpanded += 1
    if nexpanded > 0:
        log.info('    Found {0} {1}x{1} masked regions while '
                 'cleaning.'.format(nexpanded, size))
    return data


def _local_median(data_nanmask, x, y, nx, ny, size=5, expanded=False):
    """Compute the local median in a 2D window, excluding NaN."""

    hy, hx = size // 2, size // 2
    x0, x1 = np.array([x - hx, x + hx + 1]).clip(0, nx)
    y0, y1 = np.array([y - hy, y + hy + 1]).clip(0, ny)
    region = data_nanmask[y0:y1, x0:x1].ravel()
    goodpixels = region[np.isfinite(region)]
    if len(goodpixels) > 0:
        median_val = np.median(goodpixels)
    else:
        newsize = size + 2     # keep size odd
        median_val, expanded = _local_median(data_nanmask, x, y, nx, ny,
                                             size=newsize, expanded=True)
    return median_val, expanded



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
Package to remove cosmic rays from an astronomical image using the
L.A.Cosmic (PASP 113, 1420, 2001) algorithm.
"""

# Affiliated packages may add whatever they like to this file, but
# should keep this content at the top.
# ----------------------------------------------------------------------------
from ._astropy_init import *
# ----------------------------------------------------------------------------

if not _ASTROPY_SETUP_:
    from .lacosmic import *



# Autogenerated by Astropy-affiliated package astropy_helpers's setup.py on 2016-05-09 15:26:59.132943
from __future__ import unicode_literals
import datetime

version = "1.1.2"
githash = "d8f48901442e4056879c4249e73ecd7d04a28282"


major = 1
minor = 1
bugfix = 2

release = True
timestamp = datetime.datetime(2016, 5, 9, 15, 26, 59, 132943)
debug = False

try:
    from ._compiler import compiler
except ImportError:
    compiler = "unknown"

try:
    from .cython_version import cython_version
except ImportError:
    cython_version = "unknown"



# Licensed under a 3-clause BSD style license - see LICENSE.rst


def get_package_data():
    # Install the theme files
    return {
        'astropy_helpers.sphinx': [
            'ext/templates/*/*',
            'local/*.inv',
            'themes/bootstrap-astropy/*.*',
            'themes/bootstrap-astropy/static/*.*']}



# Licensed under a 3-clause BSD style license - see LICENSE.rst
from __future__ import (absolute_import, division, print_function,
                        unicode_literals)
from distutils.version import LooseVersion
from astropy.tests.helper import pytest
import numpy as np
from numpy.testing import assert_array_equal, assert_allclose
from ..lacosmic import lacosmic

try:
    import scipy
    HAS_SCIPY = True
except ImportError:
    HAS_SCIPY = False

import astropy
if LooseVersion(astropy.__version__) < LooseVersion('1.1'):
    ASTROPY_LT_1P1 = True
else:
    ASTROPY_LT_1P1 = False

IMG = np.random.RandomState(1234567890).randn(5, 5) * 0.5
x, y = [0, 3, 4, 2, 1], [0, 2, 1, 4, 2]
CR = np.zeros(IMG.shape)
CR[y, x] = 10.
MASK_REF = CR.astype(np.bool)
CR_IMG = IMG + CR


@pytest.mark.skipif('not HAS_SCIPY')
@pytest.mark.skipif('ASTROPY_LT_1P1')
class TestLACosmic(object):
    def test_lacosmic(self):
        """Test basic lacosmic."""
        crclean_img, crmask_img = lacosmic(CR_IMG, 3, 2, 2, effective_gain=1,
                                           readnoise=0)
        assert_allclose(crclean_img, IMG, atol=0.76)
        assert_array_equal(crmask_img, MASK_REF)

    def test_background_scalar(self):
        """Test lacosmic with a background scalar."""
        crclean_img, crmask_img = lacosmic(CR_IMG, 3, 1, 1, effective_gain=1,
                                           readnoise=0, background=10)
        assert_allclose(crclean_img, IMG, atol=0.76)
        assert_array_equal(crmask_img, MASK_REF)

    def test_background_maxiter(self):
        """Test lacosmic with a background scalar."""
        crclean_img, crmask_img = lacosmic(CR_IMG, 3, 1, 1, effective_gain=1,
                                           readnoise=0, background=10,
                                           maxiter=1)
        assert_allclose(crclean_img, IMG, atol=0.76)
        assert_array_equal(crmask_img, MASK_REF)

    def test_background_image(self):
        """Test lacosmic with a 2D background image."""
        bkgrd_img = np.ones(IMG.shape) * 10.
        crclean_img, crmask_img = lacosmic(CR_IMG, 3, 1, 1, effective_gain=1,
                                           readnoise=0, background=bkgrd_img)
        assert_allclose(crclean_img, IMG, atol=0.76)
        assert_array_equal(crmask_img, MASK_REF)

    def test_mask_image(self):
        """Test lacosmic with an input mask image."""
        mask = MASK_REF.copy()
        mask[2, 1] = False
        mask_ref2 = np.logical_and(MASK_REF, ~mask)
        crclean_img, crmask_img = lacosmic(CR_IMG, 3, 2, 2, effective_gain=1,
                                           readnoise=0, mask=mask)
        assert_allclose(crclean_img * mask_ref2, IMG * mask_ref2, atol=0.76)
        assert_array_equal(crmask_img, mask_ref2)

    def test_error_image(self):
        """Test lacosmic with an input error image."""
        error = np.sqrt(IMG.clip(min=0.001))
        crclean_img, crmask_img = lacosmic(CR_IMG, 3, 2, 2, effective_gain=1,
                                           readnoise=0, error=error)
        assert_allclose(crclean_img, IMG, atol=0.76)
        assert_array_equal(crmask_img, MASK_REF)

    def test_large_cosmics(self):
        """Test lacosmic cleaning with large cosmic rays."""
        test_img = np.ones((7, 7))
        test_img[1:6, 1:6] = 100.
        mask_ref2 = np.zeros((7, 7), dtype=np.bool)
        mask_ref2[1:6, 1:6] = True
        crclean_img, crmask_img = lacosmic(test_img, 3, 2, 2, effective_gain=1,
                                           readnoise=0)
        assert_array_equal(crmask_img, mask_ref2)

    def test_error_image_size(self):
        """
        Test if AssertionError raises if shape of error doesn't match
        image.
        """
        error = np.zeros((7, 7))
        with pytest.raises(ValueError):
            lacosmic(CR_IMG, 3, 2, 2, effective_gain=1, readnoise=0,
                     error=error)
